#define _GNU_SOURCE
#include "nfs_config.h"

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <mntent.h>
#include <errno.h>
#include <string.h>
#include <sys/vfs.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <syslog.h>
#include <uv.h>
#include <glib/gi18n.h>
#include <locale.h>

#include "list.h"

#define FILE_SIZE_KB 0x400
#define FILE_SIZE_MB 0x100000
#define FILE_SIZE_GB 0x40000000
#define TMP_FILE ".dwp.db"
#define PROC_MOUNTS "/proc/mounts"

#define GETTEXT_PACKAGE "nfs-check-disk"

typedef struct disk_info
{
  char *mount_point;
  char *fsname;
  char *type;
  fsblkcnt_t tital_size;
  fsblkcnt_t free_size;
  struct list_head disk_node;
} disk_info, *disk_info_p;

static LIST_HEAD(ext4_disk);

static const char *pidfile = "/var/run/diskprotect.pid";
uv_timer_t *handle;

static int write_pid_file(void)
{
  int pidfd, len;
  char val[16];

  len = snprintf(val, sizeof(val), "%u\n", getpid());
  if (len < 0)
  {
    syslog(LOG_ERR, "Pid error (%s)", strerror(errno));
    pidfile = 0;
    return 1;
  }
  pidfd = open(pidfile, O_CREAT | O_TRUNC | O_NOFOLLOW | O_WRONLY, 0644);
  if (pidfd < 0)
  {
    syslog(LOG_ERR, "Unable to set pidfile (%s)", strerror(errno));
    pidfile = 0;
    return 1;
  }
  if (write(pidfd, val, (unsigned int)len) != len)
  {
    syslog(LOG_ERR, "Unable to write to pidfile (%s)", strerror(errno));
    close(pidfd);
    return 1;
  }
  close(pidfd);
  return 0;
}

disk_info_p create_disk_info()
{
  disk_info_p d = malloc(sizeof(disk_info));
  memset(d, 1, sizeof(disk_info));

  return d;
}

void del_disk_info(disk_info_p info)
{
  if (!info)
  {
    return;
  }
  if (info->mount_point)
  {
    free(info->mount_point);
    info->mount_point = NULL;
  }
  if (info->fsname)
  {
    free(info->fsname);
    info->fsname = NULL;
  }
  if (info->type)
  {
    free(info->type);
    info->type = NULL;
  }
  free(info);
}

off_t get_tmp_file_size(char *path)
{
  int rec = 0;
  char tmp_file[4096] = {
      '\0',
  };
  struct stat *buf;

  sprintf(tmp_file, "%s/%s", path, TMP_FILE);
  rec = access(tmp_file, F_OK);
  if (rec < 0)
  {
    return rec;
  }

  buf = (struct stat *)malloc(sizeof(struct stat));
  rec = stat(tmp_file, buf);

  if (rec < 0)
  {
    return -1;
  }

  return buf->st_size;
}

static void configDisk()
{
  struct statfs fs;
  disk_info_p entry, n;
  list_for_each_entry_safe(entry, n, &ext4_disk, disk_node)
  {
    if (statfs(entry->mount_point, &fs) < 0)
    {
      del_disk_info(entry);
      continue;
    }
    char command[256] = {
        '\0',
    };
    char buffer[2048] = {
        '\0',
    };
    sprintf(command, "tune2fs -m 0 %s", entry->fsname);
    FILE *pfd = popen(command, "r");
    fread(buffer, 1, 2048, pfd);
    int status = pclose(pfd);
    if (!WIFEXITED(status))
    {
      printf("set %s reserved blocks error: exit %d %s\n", entry->fsname, WEXITSTATUS(status), buffer);
    }
  }
}

int find_mounted_ext4_partition(struct list_head *list)
{
  char *filename = PROC_MOUNTS;
  FILE *mntfile;
  struct mntent *mntent;

  mntfile = setmntent(filename, "r");
  if (!mntfile)
  {
    syslog(LOG_ERR, "Failed to read mtab file, error [%s]\n",
           strerror(errno));
    return -1;
  }
  while ((mntent = getmntent(mntfile)))
  {
    if (strncmp(mntent->mnt_type, "ext4", 4) == 0 && strncmp(mntent->mnt_dir, "/boot", 5))
    {
      disk_info_p entry = create_disk_info();
      entry->mount_point = strdup(mntent->mnt_dir);
      entry->fsname = strdup(mntent->mnt_fsname);
      entry->type = strdup(mntent->mnt_type);
      list_add_tail(&entry->disk_node, list);
    }
  }

  endmntent(mntfile);
  return 0;
}

int realloca_tmp_file(char *path, off_t file_size)
{
  int rec = 0;

  char tmp_file[4096] = {
      '\0',
  };

  sprintf(tmp_file, "%s/%s", path, TMP_FILE);

  int fd = open(tmp_file, O_CREAT | O_RDWR | O_TRUNC, 0700);

  rec = posix_fallocate(fd, 0, file_size);

  return rec;
}

int disk_space_check(int login_clear)
{
  long int rec;
  struct statfs fs;
  long long seat_size = 0;

  disk_info_p entry, n;
  list_for_each_entry_safe(entry, n, &ext4_disk, disk_node)
  {
    if (statfs(entry->mount_point, &fs) < 0)
    {
      del_disk_info(entry);
      continue;
    }
    entry->tital_size = (fs.f_blocks) * (fs.f_bsize);
    entry->free_size = (fs.f_bavail) * (fs.f_bsize);
    rec = get_tmp_file_size(entry->mount_point);
    uint64_t cutSize = 20L;
    if (login_clear)
    {
      cutSize = 2L;
    }
    if (entry->free_size + rec > 502L * FILE_SIZE_MB)
    {
      seat_size = 500 * FILE_SIZE_MB;
    }
    else if (entry->free_size + rec > cutSize * 4 * FILE_SIZE_MB)
    {
      seat_size = (entry->free_size + rec) - cutSize * FILE_SIZE_MB;
    }
    else
    {
      seat_size = (entry->free_size + rec) * 4 / 5;
    }
    // syslog(LOG_DEBUG, "entry = %s %ld %ld\n", entry->mount_point, entry->free_size, rec);
    realloca_tmp_file(entry->mount_point, seat_size);
  }

  return 0;
}

static unsigned int size2mb(long size)
{
  return (size >= FILE_SIZE_MB ? (unsigned int)(size >> 20) : 0);
}

static unsigned int size2kb(long size)
{
  return (size >= FILE_SIZE_MB ? (unsigned int)(size >> 10) : 0);
}

static char *disk_space_remind()
{

  struct statfs fs;
  long long free_size = 0;
  disk_info_p entry, n;
  list_for_each_entry_safe(entry, n, &ext4_disk, disk_node)
  {
    if (statfs(entry->mount_point, &fs) < 0)
    {
      continue;
    }

    free_size = fs.f_bavail * fs.f_bsize;

    // printf("%d\n", size2mb(free_size));
    if (free_size < 2L * FILE_SIZE_GB)
    {
      char command[256] = {
          '\0',
      };
      if (free_size > 2L * FILE_SIZE_MB)
      {
        sprintf(command, "wall -n \"%s %s %s (%s %dMB), %s !!!\"", _("Warnning"), entry->mount_point, _("Low Disk Space on"), _("below"), size2mb(free_size), _("Please clean immediattely"));
      }
      else
      {
        // sprintf(command, "wall -n\"Warnning: Low Disk Space on %s (below %dkb), Please clean immediattely !\"", entry->mount_point, size2kb(free_size));
        sprintf(command, "wall -n \"%s %s %s (%s %dkb), %s !!!\"", _("Warnning"), entry->mount_point, _("Low Disk Space on"), _("below"), size2kb(free_size), _("Please clean immediattely"));
      }
      system(command);
      memset(command, 0, 256);
      if (free_size > 2L * FILE_SIZE_MB)
      {
        sprintf(command, "%s %s %s (%s %dMB), %s !!!", _("Warnning"), entry->mount_point, _("Low Disk Space on"), _("below"), size2mb(free_size), _("Please clean immediattely"));
      }
      else
      {
        sprintf(command, "%s %s %s (%s %dkb), %s !!!", _("Warnning"), entry->mount_point, _("Low Disk Space on"), _("below"), size2kb(free_size), _("Please clean immediattely"));
      }
      syslog(LOG_INFO, "%s", command);
      continue;
    }
    else
    {
      continue;
    }
  }

  return 0;
}

static void check_timer()
{
  printf("run check!\n");
  disk_space_remind();
}

uv_loop_t *loop;

void sig_hdlr(__attribute__((unused)) int signo, __attribute__((unused)) siginfo_t *info, __attribute__((unused)) void *act)
{
  uv_stop(loop);
  uv_loop_close(loop);
}

int main(int argc, char **argv)
{
  openlog("nfs_check_disk", LOG_CONS | LOG_PID, LOG_DAEMON);
  INIT_LIST_HEAD(&ext4_disk);
  find_mounted_ext4_partition(&ext4_disk);

  bindtextdomain(GETTEXT_PACKAGE, "/usr/share/locale");
  bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
  textdomain(GETTEXT_PACKAGE);
  setlocale(LC_ALL, "");

  disk_config *config = read_conf();
  if (config->set_reserved_to_zero)
  {
    configDisk();
    config->set_reserved_to_zero = 0;
    rewrite_config();
  }

  int login_clear = 0;

  if (argc > 1 && !strncmp(argv[1], "check_once", strlen("check_once")))
  {
    login_clear = 1;
  }

  disk_space_check(login_clear);

  if (argc > 1 && !strncmp(argv[1], "check_once", strlen("check_once")))
  {
    closelog();
    return 0;
  }

  if (daemon(0, 0) < 0)
  {
    closelog();
    exit(-1);
  }
  printf("start daemon\n");

  write_pid_file();

  struct sigaction act;
  memset(&act, 0, sizeof(act));
  act.sa_flags = SA_SIGINFO;
  act.sa_sigaction = sig_hdlr;
  if (sigaction(SIGSEGV, &act, NULL) == -1 ||
      sigaction(SIGINT, &act, NULL) == -1 ||
      sigaction(SIGTSTP, &act, NULL) == -1 ||
      sigaction(SIGTERM, &act, NULL) == -1)
  {
    printf("Install sig SIGSEGV failed : %s\n", strerror(errno));
  }

  // loop = malloc(sizeof(uv_loop_t));
  loop = uv_default_loop();
  uv_loop_init(loop);
  uv_timer_t uv_timer;
  uv_timer_init(loop, &uv_timer);
  uv_timer_start(&uv_timer, check_timer, 10 * 1000, 180 * 1000);

  uv_run(loop, UV_RUN_DEFAULT);

  free(loop);

  disk_info_p entry, n;
  list_for_each_entry_safe(entry, n, &ext4_disk, disk_node)
  {
    del_disk_info(entry);
  }
  closelog();

  return 0;
}
